"""
This module contains the restrictions that can be applied or checked in the simulation environment.
"""

# flake8: noqa: F401

from pydistsim.restrictions.base_restriction import (
    ApplicableRestriction,
    CheckableRestriction,
    Restriction,
)
